/************************************************************
 **
 ** jal_asm.c : JAL assembler definitions
 **
 ** Copyright (c) 2005, Kyle A. York
 ** All rights reserved
 **
 ************************************************************/
#include "../libpic12/pic_inst.h"
#include "../libcore/cmd_asm.h"
#include "../libcore/pf_msg.h"
#include "../libcore/pf_proc.h"
#include "../libcore/pf_cmd.h"
#include "jal_expr.h"
#include "jal_tokn.h"
#include "jal_incl.h"
#include "jal_asm.h"

/* needed MACROS:
   ADDCR  f,d : btfsc _status, _c;  incf f,d
   ADDDCF f,d : btfsc _status, _dc; incf f,d
   B      n   : goto n
   BC     n   : btfsc _status, _c;  goto n
   BDC    n   : btfsc _status, _dc; goto n
   BNC    n   : btfss _status, _c;  goto n
   BNDC   n   : btfss _status, _dc; goto n
   BNZ    n   : btfss _status, _z;  goto n
   BZ     n   : btfss _status, _z;  goto n
   CLRC       : bcf _status, _c
   CLRDC      : bcf _status, _dc
   CLRZ       : bcf _status, _z
   LCALL  n   : page call n
   LGOTO  n   : page goto n
   MOVFW  f   : movf f,w
   NEGF   f,d : comf f,f; inc f,f
   SETC       : bsf _status, _c
   SETDC      : bsf _status, _dc
   SETZ       : bsf _status, _z
   SKPC       : btfss _status, _c
   SKPDC      : btfss _status, _dc
   SKPNC      : btfsc _status, _c
   SKPNDC     : btfsc _status, _dc
   SKPNZ      : btfsc _status, _z
   SKPZ       : btfss _status, _z
   SUBCF  f,d : btfsc _status, _c ; decf f,d
   SUBDCF f,d : btfsc _status, _dc; decf f,d
*/
/* there are only six types of instructions:
     op     --
     op f   --
     op f,d -- d = 0 or 1
     op f,b -- 0 <= b <= 7
     op n   -- 0 <= n <= 255
     op k   -- 0 <= k <= 2047
   plus two special commands:
     bank -- set the bank bits for the next instruction
     page -- set the page bits for the next instruction
*/
typedef enum {
  asm_cmd_special_none,
  asm_cmd_special_movfw, /* mov f,0 */
  asm_cmd_special_skpc,  /* btfss _status, _c */
  asm_cmd_special_skpnc, /* btfsc _status, _c */
  asm_cmd_special_skpz,  /* btfss _status, _z */
  asm_cmd_special_skpnz  /* btfsc _status, _z */
} asm_cmd_special_t;

static const struct {
  const char       *tag;
  pic_optype_t      type;
  pic_opcode_t      op;
  asm_cmd_special_t special; /* a hack for the special operations */
} asm_cmds[] = {
  {"addwf",  pic_optype_f_d,    pic_opcode_addwf},
  {"andwf",  pic_optype_f_d,    pic_opcode_andwf},
  {"clrf",   pic_optype_f,      pic_opcode_clrf},
  {"clrw",   pic_optype_none,   pic_opcode_clrw},
  {"comf",   pic_optype_f_d,    pic_opcode_comf},
  {"decf",   pic_optype_f_d,    pic_opcode_decf},
  {"decfsz", pic_optype_f_d,    pic_opcode_decfsz},
  {"incf",   pic_optype_f_d,    pic_opcode_incf},
  {"incfsz", pic_optype_f_d,    pic_opcode_incfsz},
  {"iorwf",  pic_optype_f_d,    pic_opcode_iorwf},
  {"movf",   pic_optype_f_d,    pic_opcode_movf},
  {"movwf",  pic_optype_f,      pic_opcode_movwf},
  {"nop",    pic_optype_none,   pic_opcode_nop},
  {"rlf",    pic_optype_f_d,    pic_opcode_rlf},
  {"rrf",    pic_optype_f_d,    pic_opcode_rrf},
  {"subwf",  pic_optype_f_d,    pic_opcode_subwf},
  {"swapf",  pic_optype_f_d,    pic_opcode_swapf},
  {"xorwf",  pic_optype_f_d,    pic_opcode_xorwf},
  {"bcf",    pic_optype_f_b,    pic_opcode_bcf},
  {"bsf",    pic_optype_f_b,    pic_opcode_bsf},
  {"btfsc",  pic_optype_f_b,    pic_opcode_btfsc},
  {"btfss",  pic_optype_f_b,    pic_opcode_btfss},
  {"addlw",  pic_optype_n,      pic_opcode_addlw},
  {"andlw",  pic_optype_n,      pic_opcode_andlw},
  {"call",   pic_optype_k,      pic_opcode_call},
  {"clrwdt", pic_optype_none,   pic_opcode_clrwdt},
  {"goto",   pic_optype_k,      pic_opcode_goto},
  {"iorlw",  pic_optype_n,      pic_opcode_iorlw},
  {"movlw",  pic_optype_n,      pic_opcode_movlw},
  {"retfie", pic_optype_none,   pic_opcode_retfie},
  {"retlw",  pic_optype_n,      pic_opcode_retlw},
  {"return", pic_optype_none,   pic_opcode_ret},
  {"sleep",  pic_optype_none,   pic_opcode_sleep},
  {"sublw",  pic_optype_n,      pic_opcode_sublw},
  {"xorlw",  pic_optype_n,      pic_opcode_xorlw},
  {"option", pic_optype_none,   pic_opcode_option}, /* don't use! */
  {"tris",   pic_optype_tris,   pic_opcode_tris},   /* don't use! */
  /* movfw looks like it's supposed to be : movf x,0 */
  {"movfw",  pic_optype_f,      pic_opcode_movf, asm_cmd_special_movfw},
  /* let's add some common macros */
  {"skpc",    pic_optype_macro, pic_opcode_none, asm_cmd_special_skpc},
  {"skpnc",   pic_optype_macro, pic_opcode_none, asm_cmd_special_skpnc},
  {"skpz",    pic_optype_macro, pic_opcode_none, asm_cmd_special_skpz},
  {"skpnz",   pic_optype_macro, pic_opcode_none, asm_cmd_special_skpnz},
};

/* parse a single assembly command */
#define ASM_BIT_BANK 0x0001 /* BANK found */
#define ASM_BIT_PAGE 0x0002 /* PAGE found */

static void asm_cmd_add(pfile_t *pf,
    pic_opcode_t op, value_t val, pic_opdst_t opdst, value_t n,
    label_t lbl, flag_t flags)
{
  cmd_t cmd;

  cmd = cmd_asm_alloc(op, val, opdst, n, lbl, flags);
  if (cmd) {
    pfile_cmd_add(pf, cmd);
  }
}

/* this returns TRUE if at least one token has been absorbed */
void jal_parse_asm(pfile_t *pf)
{
  unsigned     flags;
  size_t       ii;
  value_t      val;
  value_t      n;
  pic_opcode_t op;
  pic_opdst_t  opdst;
  label_t      lbl;
  boolean_t    err;
  unsigned     n_lo;
  unsigned     n_hi;
 
  err   = boolean_false;
  flags = 0;
  op    = pic_opcode_none;
  opdst = pic_opdst_none;
  val   = VALUE_NONE;
  n     = VALUE_NONE;
  lbl   = LABEL_NONE;
  n_lo  = 0;
  n_hi  = -1;

  /* first, look for any pre BANK or PAGE elements */
  /* it looks like the assembler should handle the following:
     BANK value
     PAGE label
     BANK | PAGE statement
     Logically, only one of BANK or PAGE can exist for a statement and
     the flag effects only the *following* statement */
  if (pf_token_is(pf, pf_token_current, pfile_log_none, "bank")) {
    flags |= ASM_BIT_BANK;
    pf_token_get(pf, pf_token_next);
  } else if (pf_token_is(pf, pf_token_current, pfile_log_none, "page")) {
    flags |= ASM_BIT_PAGE;
    pf_token_get(pf, pf_token_next);
  }
  for (ii = 0; 
       (ii < COUNT(asm_cmds)) 
       && !pf_token_is(pf, pf_token_current, pfile_log_none, 
         asm_cmds[ii].tag);
       ii++)
    ; /* null body */


  if (COUNT(asm_cmds) == ii) {
    /* let's allow:
       PAGE label
       BANK [# | value] */
    if (flags & ASM_BIT_BANK) {
      /* apparently there's some assembly command ``bank n'' */
      /* so allow it here */
      val = pfile_value_find(pf, pfile_log_none,
          pf_token_get(pf, pf_token_current));
      if (val && !value_is_const(val)) {
        pf_token_get(pf, pf_token_next);
      } else {
        if (val) {
          value_release(val);
        }
        val = jal_parse_expr(pf);
        if (val) {
          if (!value_is_const(val)) {
            pfile_log(pf, pfile_log_err, PFILE_MSG_CONSTANT_EXPECTED);
            err = boolean_true;
          }
          value_release(val);
          val = VALUE_NONE;
          flags &= ~ASM_BIT_BANK;
        }
      }
    } else if (flags & ASM_BIT_PAGE) {
      lbl = pfile_label_find(pf, pfile_log_err, 
              pf_token_get(pf, pf_token_current));
      if (lbl) {
        pf_token_get(pf, pf_token_next);
      } else {
        err = boolean_true;
      }
    } else {
      err = boolean_true;
      pfile_log(pf, pfile_log_err, PFILE_MSG_SYNTAX_ERROR);
    }
  } else {
    /* process the cmd */
    op = asm_cmds[ii].op;
    pf_token_get(pf, pf_token_next);

    switch (asm_cmds[ii].type) {
      case pic_optype_none:
        /* nothing more to do */
        break;
      case pic_optype_f:
      case pic_optype_f_d:
      case pic_optype_f_b:
        /* allow ''BANK'', a variable name, or a constant expression */
        if (pf_token_is(pf, pf_token_current, pfile_log_none, "bank")) {
          flags |= ASM_BIT_BANK;
          pf_token_get(pf, pf_token_next);
        }
        val = pfile_value_find(pf, pfile_log_none,
            pf_token_get(pf, pf_token_current));
        if (val) {
          pf_token_get(pf, pf_token_next);
          if (value_is_array(val)
              && (pf_token_is(pf, pf_token_current, pfile_log_none, "["))) {
            value_t               subs;

            pf_token_get(pf, pf_token_next);
            subs = jal_parse_expr(pf);
            if (subs) {
              variable_def_t        def;
              variable_def_member_t mbr;

              mbr = variable_def_member_get(
                      value_def_get(val));
              def = variable_def_member_def_get(mbr);

              if (!value_is_const(subs)) {
                pfile_log(pf, pfile_log_err, PFILE_MSG_CONSTANT_EXPECTED);
              } else if (value_const_get(subs) 
                  >= variable_def_member_ct_get(mbr)) {
                pfile_log(pf, pfile_log_err, "subscript out of range");
              } else {
                if (variable_def_sz_get(def) > 1) {
                  value_t tmp;

                  tmp = pfile_constant_get(pf, variable_def_sz_get(def)
                      * value_const_get(subs), VARIABLE_DEF_NONE);
                  value_release(subs);
                  subs = tmp;
                }
                value_baseofs_set(val, subs);
              }
              value_release(subs);
            }
            if (pf_token_is(pf, pf_token_current, pfile_log_err, "]")) {
              pf_token_get(pf, pf_token_next);
            }
          }
        } else {
          val = jal_parse_expr(pf);
          if (val && !value_is_const(val)) {
            pfile_log(pf, pfile_log_err, PFILE_MSG_CONSTANT_EXPECTED);
            value_release(val);
            val = VALUE_NONE;
            err = boolean_true;
          }
        }
        if (pic_optype_f == asm_cmds[ii].type) {
          if (asm_cmd_special_movfw == asm_cmds[ii].special) {
            n = pfile_value_find(pf, pfile_log_none, "w");
            n_lo = 0;
            n_hi = 1;
          }
        } else if (pic_optype_f_d == asm_cmds[ii].type) {
          if (pf_token_is(pf, pf_token_current, pfile_log_err, ",")) {
            pf_token_get(pf, pf_token_next);
          }
          n = jal_parse_expr(pf);
          n_lo = 0;
          n_hi = 1;
        } else {
          /* if value is a bit variable, position is assumed */
          if (!value_dflag_test(val, VARIABLE_DEF_FLAG_BIT)) {
            if (pf_token_is(pf, pf_token_current, pfile_log_err, ",")) {
              pf_token_get(pf, pf_token_next);
            }
            n = jal_parse_expr(pf);
            n_lo = 0;
            n_hi = 7;
          }
        }
        break;
      case pic_optype_n:
        n = jal_parse_expr(pf);
        n_lo = 0;
        n_hi = 255;
        break;
      case pic_optype_tris:
        n = jal_parse_expr(pf);
        n_lo = 5;
        n_hi = 7;
        break;
      case pic_optype_k:
        if (pf_token_is(pf, pf_token_current, pfile_log_none, "page")) {
          flags |= ASM_BIT_PAGE;
          pf_token_get(pf, pf_token_next);
        }
        lbl = pfile_label_find(pf, pfile_log_none, pf_token_ptr_get(pf));
        if (lbl) {
          pf_token_get(pf, pf_token_next);
        } else {
          n = jal_parse_expr(pf);
          if ((VALUE_NONE == n) || !value_is_const(n)) {
            pfile_log(pf, pfile_log_err, "constant expression expected");
            value_release(n);
            n = VALUE_NONE;
            err = boolean_true;
          } else {
            lbl = pfile_label_alloc(pf, 0);
            label_pc_set(lbl, value_const_get(n));
            label_flag_set(lbl, LABEL_FLAG_DEFINED);
            label_flag_set(lbl, LABEL_FLAG_USER);
            value_release(n);
            n = VALUE_NONE;
          }
        }
        break;
      case pic_optype_macro:
        {
          const char  *flag;

          flag = 0;
          switch (asm_cmds[ii].special) {
            case asm_cmd_special_skpc:  
              flag = "_c"; 
              op   = pic_opcode_btfss; 
              break;
            case asm_cmd_special_skpnc: 
              flag = "_c"; 
              op   = pic_opcode_btfsc; 
              break;
            case asm_cmd_special_skpz:  
              flag = "_z"; 
              op   = pic_opcode_btfss; 
              break;
            case asm_cmd_special_skpnz: 
              flag = "_z"; 
              op   = pic_opcode_btfsc; 
              break;
            case asm_cmd_special_none:
            case asm_cmd_special_movfw:
              break;
          }

          val = pfile_value_find(pf, pfile_log_err, "_status");
          n   = pfile_value_find(pf, pfile_log_err, flag);
        }
        break;
    }
  }
  if (n) {
    if (!value_is_const(n)) {
      pfile_log(pf, pfile_log_err, PFILE_MSG_CONSTANT_EXPECTED);
      err = boolean_true;
    } else if ((value_const_get(n) < n_lo) 
      || (value_const_get(n) > n_hi)) {
      pfile_log(pf, pfile_log_err, PFILE_MSG_CONSTANT_RANGE);
      err = boolean_true;
    }
    if (err) {
      value_release(n);
      n = VALUE_NONE;
    }
  }

  if (!err) {
    if (flags & ASM_BIT_PAGE) {
      if (!lbl) {
        pfile_log(pf, pfile_log_warn, 
            "PAGE used without an associated label");
      } else {
        /* generate page bits */
        value_t code_sz;

        code_sz = pfile_value_find(pf, pfile_log_err, "_code_size");
        if (value_const_get(code_sz) > 2048) {
          asm_cmd_add(pf, pic_opcode_branchlo_set, VALUE_NONE, pic_opdst_none,
              VALUE_NONE, lbl, pfile_proc_flag_test(pfile_proc_active_get(pf),
              PFILE_PROC_FLAG_KEEP_PAGE) ? PIC_CODE_FLAG_NO_OPTIMIZE : 0);
        }
        if (value_const_get(code_sz) > 4096) {
          asm_cmd_add(pf, pic_opcode_branchhi_set, VALUE_NONE, pic_opdst_none,
              VALUE_NONE, lbl, pfile_proc_flag_test(pfile_proc_active_get(pf),
              PFILE_PROC_FLAG_KEEP_PAGE) ? PIC_CODE_FLAG_NO_OPTIMIZE : 0);
        }
        value_release(code_sz);
      }
    }
    if (flags & ASM_BIT_BANK) {
      if (!val) {
        pfile_log(pf, pfile_log_warn,
            "BANK used without an associated value");
      } else {
        asm_cmd_add(pf, pic_opcode_datalo_set, val, pic_opdst_none,
            VALUE_NONE, LABEL_NONE, 
            pfile_proc_flag_test(pfile_proc_active_get(pf),
              PFILE_PROC_FLAG_KEEP_BANK) ? PIC_CODE_FLAG_NO_OPTIMIZE : 0);
        asm_cmd_add(pf, pic_opcode_datahi_set, val, pic_opdst_none,
            VALUE_NONE, LABEL_NONE,
            pfile_proc_flag_test(pfile_proc_active_get(pf),
              PFILE_PROC_FLAG_KEEP_BANK) ? PIC_CODE_FLAG_NO_OPTIMIZE : 0);
      }
    }
    if ((pic_optype_f_d == asm_cmds[ii].type)
      || (asm_cmd_special_movfw == asm_cmds[ii].special)) {
      opdst = value_const_get(n) ? pic_opdst_f : pic_opdst_w;
      value_release(n);
      n = VALUE_NONE;
    }

    asm_cmd_add(pf, op, val, opdst, n, lbl, PIC_CODE_FLAG_NO_OPTIMIZE);
    label_release(lbl);
    value_release(n);
    value_release(val);
  }
}

void jal_parse_assembler(pfile_t *pf)
{
  pfile_pos_t pos_start;

  pf_token_start_get(pf, &pos_start);
  pfile_block_enter(pf);
  while (!pf_token_is_eof(pf)
    && !pf_token_is(pf, pf_token_current, pfile_log_none, "end")) {
    pfile_token_ct_t token_ct;
    const char      *ptr;

    token_ct = pfile_token_ct_get(pf);
    pf_token_start_get(pf, &pos_start);
    pfile_statement_start_set(pf, &pos_start);
    ptr = pf_token_get(pf, pf_token_current);
    if (pf_token_is(pf, pf_token_current, pfile_log_none, "pragma")) {
      /* allow pragmas in assembler blocks. who knew? */
      pf_token_get(pf, pf_token_next);
      jal_parse_pragma(pf);
    } else if (pf_token_is(pf, pf_token_current, pfile_log_none, "local")) {
      /* declare some local labels */
      do {
        ptr = pf_token_get(pf, pf_token_next);
        if (jal_token_is_identifier(pf)) {
          label_release(pfile_label_alloc(pf, ptr));
          pf_token_get(pf, pf_token_next);
        }
      } while (pf_token_is(pf, pf_token_current, pfile_log_none, ","));
    } else {
      /* maybe this is a label definition */
      label_t lbl;

      lbl = LABEL_NONE;
      if (jal_token_is_identifier(pf)) {
        lbl = pfile_label_find(pf, pfile_log_none, ptr);
        if (lbl) {
          pfile_cmd_label_add(pf, lbl);
          label_release(lbl);
          if (pf_token_is(pf, pf_token_next, pfile_log_err, ":")) {
            pf_token_get(pf, pf_token_next);
          }
        }
      }
      if (!lbl) {
        jal_parse_asm(pf);
      }
    }
    /* if nothing has been parsed we need to skip the current token */
    if (token_ct == pfile_token_ct_get(pf)) {
      pf_token_get(pf, pf_token_next);
    }
  }
  if (pf_token_is(pf, pf_token_next, pfile_log_err, "assembler")) {
    pf_token_get(pf, pf_token_next);
  }
  pfile_block_leave(pf);
  pf_token_start_get(pf, &pos_start);
  pfile_statement_start_set(pf, &pos_start);
}

